import numpy as np
import torch
from scipy import spatial
# import pytest

from ezyrb import POD, RBF, Database, Snapshot, Parameter, Linear, ANN
from ezyrb import ReducedOrderModel as ROM
from ezyrb.plugin import AutomaticShiftSnapshots

n_params = 15
params = np.linspace(0.5, 4.5, n_params).reshape(-1, 1)


def gaussian(x, mu, sig):
    return np.exp(-np.power(x - mu, 2.) / (2 * np.power(sig, 2.)))


def wave(t, res=256):
    x = np.linspace(0, 5, res)
    return x, gaussian(x, t, 0.1)


def test_constructor():
    interp = ANN([10, 10], torch.nn.Softplus, 1e-4)
    shift = ANN([10, 10], torch.nn.Softplus, 1e-4)
    AutomaticShiftSnapshots(shift, interp, RBF())


"""
def test_fit_train():
    seed = 1
    torch.manual_seed(seed)
    np.random.seed(seed)
    interp = ANN([10, 10], torch.nn.Softplus(), 10000, frequency_print=200, lr=0.03)
    shift  = ANN([], torch.nn.LeakyReLU(), [1e-4, 5000], frequency_print=200, l2_regularization=0,  lr=0.0023)
    nnspod = AutomaticShiftSnapshots(shift, interp, Linear(fill_value=0.0), barycenter_loss=10.)
    pod = POD(rank=1)
    rbf = RBF()
    db = Database()
    for param in params:
        space, values = wave(param)
        snap = Snapshot(values=values, space=space)
        db.add(Parameter(param), snap)

    rom = ROM(db, pod, rbf, plugins=[nnspod])
    rom.fit()

    pred = rom.predict(db)

    error = 0.0
    for (_, snap), (_, truth_snap) in zip(pred._pairs, db._pairs):
        tree = spatial.KDTree(truth_snap.space.reshape(-1, 1))
        for coord, value in zip(snap.space, snap.values):
            a = tree.query(coord)
            error += np.abs(value - truth_snap.values[a[1]])

    assert error < 100.
"""

###################### TODO: extremely long test, need to rethink it
# def test_fit_test():
#     interp = ANN([10, 10], torch.nn.Softplus(), 1000, frequency_print=200, lr=0.03)
#     shift = ANN([], torch.nn.LeakyReLU(), [1e-3, 4000], optimizer=torch.optim.Adam, frequency_print=200, l2_regularization=0,  lr=0.0023)
#     nnspod = AutomaticShiftSnapshots(shift, interp, Linear(fill_value=0.0), barycenter_loss=20.)
#     pod = POD(rank=1)
#     rbf = RBF()
#     db = Database()
#     for param in params:
#         space, values = wave(param)
#         snap = Snapshot(values=values, space=space)
#         db.add(Parameter(param), snap)
#     db_train, db_test = db.split([len(db)-3, 3])
#     rom = ROM(db_train, pod, rbf, plugins=[nnspod])
#     for _ in range(10):
#         rom.fit()
#         if rom.plugins[0].shift_network.loss_trend[-1] < 1e-3:
#             break
#     pred = rom.predict(db_test.parameters_matrix)
    
#     error = 0.0
#     for (_, snap), (_, truth_snap) in zip(pred._pairs, db_test._pairs):
#         tree = spatial.KDTree(truth_snap.space.reshape(-1, 1))
#         for coord, value in zip(snap.space, snap.values):
#             a = tree.query(coord)
#             error += np.abs(value - truth_snap.values[a[1]])

#     assert error < 25.



"""EZyRB package"""

__all__ = [
    'Database', 'Snapshot', 'Reduction', 'POD', 'Approximation', 'RBF', 'Linear', 'GPR',
    'ANN', 'KNeighborsRegressor', 'RadiusNeighborsRegressor', 'AE',
    'ReducedOrderModel', 'PODAE', 'RegularGrid',
    'MultiReducedOrderModel'
]

from .database import Database
from .snapshot import Snapshot
from .parameter import Parameter
from .reducedordermodel import ReducedOrderModel, MultiReducedOrderModel
from .reduction import *
from .approximation import *
from .regular_grid import RegularGrid



"""Module for the Plugin abstract class."""

from abc import ABC


class Plugin(ABC):
    """
    The abstract `Approximation` class.

    All the classes that implement the input-output mapping should be inherited
    from this class.
    """
    def fit_preprocessing(self, rom):
        """ Void """
        pass

    def fit_before_reduction(self, rom):
        """ Void """
        pass

    def fit_after_reduction(self, rom):
        """ Void """
        pass
    
    def fit_before_approximation(self, rom):
        """ Void """
        pass

    def fit_after_approximation(self, rom):
        """ Void """
        pass

    def fit_postprocessing(self, rom):
        """ Void """
        pass

    def predict_preprocessing(self, rom):
        """ Void """
        pass

    def predict_before_approximation(self, rom):
        """ Void """
        pass

    def predict_after_approximation(self, rom):
        """ Void """
        pass

    def predict_before_expansion(self, rom):
        """ Void """
        pass
    
    def predict_after_expansion(self, rom):
        """ Void """
        pass

    def predict_postprocessing(self, rom):
        """ Void """
        pass






from .plugin import Plugin
import numpy as np
from ..approximation.rbf import RBF
from ..approximation.linear import Linear
from ..database import Database
import copy
from scipy.optimize import minimize


class Aggregation(Plugin):
    """
    The `Aggregation` plugin class.
    It implements a general framework for space-dependent aggregation of
    multiple reduced order models (ROMs). The aggregation weights can be
    computed either through standard space-dependent methods (e.g., Gaussian
    functions) or by fitting a regression/interpolation model (e.g., RBF,
    linear, ANN) on the validation set.

    Reference: Ivagnes A., Tonicello N., Rozza G. (2024). `Enhancing non-intrusive
    reduced-order models with space-dependent aggregation methods`, Acta Mechanica.

    :param fit_function: a regression/interpolation model to fit the weights
        in the validation set. If None, standard space-dependent aggregation
        methods are used (default is None).
    :param predict_function: a regression/interpolation model to predict the
        weights in the test set. Default is a linear interpolation.

    Example:
    >>> from ezyrb import POD, RBF, GPR, Database, Snapshot, Parameter, Linear, ANN
    >>> from ezyrb import ReducedOrderModel as ROM
    >>> from ezyrb.plugin import Aggregation
    >>> pod = POD(rank=10)
    >>> rbf = RBF()
    >>> gpr = GPR()
    >>> db = Database(param, snapshots.T)
    >>> rom1 = ROM(db, pod, rbf)
    >>> rom2 = ROM(db, pod, gpr)
    >>> mrom = MROM({'rbf': rom1, 'gpr': rom2}, plugins=[Aggregation(
    ...     fit_function=None, predict_function=Linear())])
    >>> mrom.fit()
    >>> pred = mrom.predict(new_param)
    """

    def __init__(self, fit_function=None, predict_function=Linear()):
        super().__init__()
        self.fit_function = fit_function
        self.predict_function = predict_function

    def _check_sum_gaussians(self, mrom, sum_gaussians, gaussians):
        """
        Method to check if the sum of the Gaussian functions is
        on-zero in all the points of the validation set.
        If there are zero values, the same weight/gaussian is taken for all the ROMs
        at those points.

        :param mrom: the multi reduced order model.
        :param sum_gaussians: the sum of the Gaussian functions.
        :param gaussians: the Gaussian functions.

        :return: the corrected Gaussian functions.
        """
        zero_values = sum_gaussians < 1e-5 # tolerance to avoid numerical issues
       
        if zero_values.any():
            gaussians[:, zero_values] = 1/len(mrom.roms) # equal weights
        return gaussians

    def _optimize_sigma(self, mrom, sigma_range=[1e-5, 1e-2]):
        """
        Method to optimize the sigma parameter in the Gaussian functions
        through a minimization procedure on the validation set.

        :param mrom: the multi reduced order model.
        :param sigma_range: the range of values to search for the optimal sigma.

        :return: the optimal sigma value.
        """
        def obj_func(sigma):
            # compute test error of multiROM from current sigma
            g_sigma = self._compute_validation_weights(mrom, sigma, normalized=True)
            mrom_prediction = np.zeros_like(mrom.validation_full_database.snapshots_matrix)
            for i, rom in enumerate(mrom.roms):
                mrom_prediction += g_sigma[i, ...] * mrom.roms[rom].predict(mrom.validation_full_database).snapshots_matrix
            test_error = np.mean(
                np.linalg.norm(mrom_prediction - mrom.validation_full_database.snapshots_matrix,
            axis=1) /  np.linalg.norm(mrom.validation_full_database.snapshots_matrix, axis=1))
            return test_error
        # minimization procedure
        res = minimize(obj_func, x0=sigma_range[0],
               method="L-BFGS-B", bounds=[sigma_range])
        print('Optimal sigma value in weights: ', res.x)
        return res.x

    def _compute_validation_weights(self, mrom, sigma, normalized=False):
        """
        Method to compute the weights in the validation set
        through standard space-dependent aggregation methods.

        :param mrom: the multi reduced order model.
        :param sigma: the sigma parameter in the Gaussian functions.
        :param normalized: if True, the weights are normalized to sum to 1
            at each spatial point (default is False).

        :return: the weights in the validation set.
        """
        validation_predicted = dict()
        for name, rom in mrom.roms.items():
            validation_predicted[name] = rom.predict(rom.validation_full_database).snapshots_matrix

        g = {}
        for k, v in validation_predicted.items():
            snap = rom.validation_full_database.snapshots_matrix
            g[k] = np.exp(- ((v - snap)**2)/(2 * (sigma**2)))

        g_tensor = np.array([g[k] for k in g.keys()])
        g_tensor = self._check_sum_gaussians(mrom, g_tensor.sum(axis=0), g_tensor)

        if normalized:
            g_tensor /= np.sum(g_tensor, axis=0)

        return g_tensor

    def fit_postprocessing(self, mrom):
        """
        Compute/fit the weights in the validation set.
        If fit_function is None, the weights are computed through standard
        space-dependent aggregation methods (e.g., Gaussian functions).
        If fit_function is not None, the weights are fitted through the
        provided regression/interpolation model (e.g., RBF, linear, ANN).

        :param mrom: the multi reduced order model.

        :return: None
        """
        rom = list(mrom.roms.values())[0]
        
        # concatenate params and space
        params = rom.validation_full_database.parameters_matrix
        input_list = []
        # Loop over the number of snapshots in the validation database
        for i in range(rom.validation_full_database.snapshots_matrix.shape[0]):
            # Get the space coordinates for the i-th snapshot
            space = rom.validation_full_database.get_snapshot_space(i)
            # Get the parameters for the i-th snapshot
            param = rom.validation_full_database.parameters_matrix[i, :]
            # Create the input array for the i-th snapshot
            snapshot_input = np.hstack([
                space,
                np.tile(param, (space.shape[0], 1))
            ])
            # Append the input array to the list
            input_list.append(snapshot_input)
        # Concatenate the input arrays for all snapshots
        input_ = np.vstack(input_list)

        # Fit the regression/interpolation that will be used to predict the
        # weights in the test database
        if self.fit_function is None:
            
            optimal_sigma = self._optimize_sigma(mrom)
            g_tensor = self._compute_validation_weights(mrom, optimal_sigma, normalized=False)
 
            self.predict_functions = []
            for i, rom in enumerate(mrom.roms.values()):
                g_ = g_tensor[i, ...].reshape(space.shape[0] * params.shape[0], -1)
                # do a copy of the function used to predict the weights,
                # otherwise we use the same for all the ROMs
                rom_func = copy.deepcopy(self.predict_function)
                # replace NaN with 0
                # TODO: this is a temporary fix, we should handle NaNs in a better way
                g_[np.isnan(g_)] = 1/len(mrom.roms)
                rom_func.fit(input_, g_)
                self.predict_functions.append(rom_func)
        
 
         # directly fit a regression to minimize the discrepancy between the aggregation and the snapshots
        elif self.fit_function is not None:
            snaps = rom.validation_full_database.snapshots_matrix
            self.fit_function.fit(input_, snaps.reshape(space.shape[0] * params.shape[0], 1))


    def predict_postprocessing(self, mrom):
        """
        Compute the weights in the test set and the predicted solution
        as a convex combination of the predicted solutions of the
        individual ROMs.

        :param mrom: the multi reduced order model.

        :return: None
        """

        # prepare the input for the prediction
        predict_weights = {}
        
        db = list(mrom.multi_predict_database.values())[0]
        
        input_list = []
        
        # Loop over the number of snapshots in the prediction database
        for i in range(mrom.predict_full_database.parameters_matrix.shape[0]):  
            # Get the space coordinates for the i-th snapshot
            space = mrom.train_full_database.get_snapshot_space(i)
            # Get the parameters for the i-th snapshot
            param = mrom.predict_full_database.parameters_matrix[i, :]
            # Create the input array for the i-th snapshot
            snapshot_input = np.hstack([
                space,
                np.tile(param, (space.shape[0], 1))
            ])
            # Append the input array to the list
            input_list.append(snapshot_input)
        # Concatenate the input arrays for all snapshots
        input_ = np.vstack(input_list)
        
        # initialize weights
        mrom.weights_predict = {}
        # predict weights with regression and normalize them
        # case without fit_function (where we use the standard space-dependent aggregation methods)
        if self.fit_function is None and self.predict_function is not None:
            
            gaussians_test = []
            for i, rom in enumerate(mrom.roms.values()):
                g_ = self.predict_functions[i].predict(input_)
                gaussians_test.append(g_)
            gaussians_test = np.array(gaussians_test)
            gaussians_test = gaussians_test.reshape(len(mrom.roms),
                    db.parameters_matrix.shape[0], space.shape[0])
            # normalize weights
            predict_weights = gaussians_test/np.sum(gaussians_test, axis=0)
            # compute predicted solution as convex combination of the reduced solutions
            for i, rom in enumerate(mrom.roms):
                mrom.weights_predict[rom] = predict_weights[i, ...]
                db = mrom.multi_predict_database[rom]
 
        # case with fit_function (ANN)
        elif self.fit_function is not None:
            weights = self.fit_function.predict(input_)
            # compute predicted solution as convex combination of the reduced solutions
            for i, rom in enumerate(mrom.roms):
                mrom.weights_predict[rom] = weights[..., i].reshape(
                    db.parameters_matrix.shape[0], -1)
        
        # compute the prediction
        prediction = np.sum([mrom.weights_predict[k] *
                mrom.multi_predict_database[k].snapshots_matrix for k in
                list(mrom.roms.keys())], axis=0)

        mrom.predict_full_database = Database(
            mrom.predict_full_database.parameters_matrix,
            prediction
            )


import numpy as np
# import pytest

from ezyrb import POD, RBF, Database, Snapshot, Parameter, Linear
from ezyrb import ReducedOrderModel as ROM
from ezyrb.plugin.shift import ShiftSnapshots

n_params = 15
params = np.linspace(0.5, 4.5, n_params).reshape(-1, 1)


def gaussian(x, mu, sig):
    return np.exp(-np.power(x - mu, 2.) / (2 * np.power(sig, 2.)))


def wave(t, res=256):
    x = np.linspace(0, 5, res)
    return x, gaussian(x, t, 0.1)


def shift(time):
    return time-0.5


def test_constructor():
    ShiftSnapshots(shift, RBF())


def test_fit():
    pod = POD()
    rbf = RBF()
    db = Database()
    for param in params:
        space, values = wave(param)
        snap = Snapshot(values=values, space=space)
        db.add(Parameter(param), snap)
    rom = ROM(db, pod, rbf, plugins=[
        ShiftSnapshots(shift, RBF())
    ])
    rom.fit()


def test_predict_ref():
    pod = POD()
    rbf = RBF()
    db = Database()
    for param in params:
        space, values = wave(param)
        snap = Snapshot(values=values, space=space)
        db.add(Parameter(param), snap)
    rom = ROM(db, pod, rbf, plugins=[
        ShiftSnapshots(shift, Linear(fill_value=0.0))
    ])
    rom.fit()
    pred = rom.predict(db._pairs[0][0].values)
    np.testing.assert_array_almost_equal(
        pred[0], db._pairs[0][1].values, decimal=1)


def test_predict():
    pod = POD()
    rbf = Linear()
    db = Database()
    for param in params:
        space, values = wave(param)
        snap = Snapshot(values=values, space=space)
        db.add(Parameter(param), snap)
    rom = ROM(db, pod, rbf, plugins=[
        ShiftSnapshots(shift, Linear(fill_value=0.0))
    ])
    rom.fit()
    pred_db = rom.predict(db)

    from scipy import spatial
    tree = spatial.KDTree(db._pairs[10][1].space.reshape(-1, 1))
    error = 0.0
    for coord, value in zip(pred_db._pairs[0][1].space, pred_db._pairs[0][1].values):
        a = tree.query(coord)
        error += np.abs(value - db._pairs[10][1].values[a[1]])
        

    assert error < 1.

# def test_values():
#     snap = Snapshot(test_value)
#     snap.values = test_value
#     snap = Snapshot(test_value, space=test_space)
#     with pytest.raises(ValueError):
#         snap.values = test_value[:-2]



""" Module for parameter object """
import numpy as np

class Parameter:

    def __init__(self, values):
        if isinstance(values, Parameter):
            self.values = values.values
        else:
            self.values = values

    @property
    def values(self):
        """ Get the snapshot values. """
        return self._values

    @values.setter
    def values(self, new_values):
        if np.asarray(new_values).ndim != 1:
            raise ValueError('only 1D array are usable as parameter.')

        self._values = np.asarray(new_values)


from ezyrb import Parameter
import numpy as np

import pytest

test_value = [0.5, 0.78]

def test_costructor():
    param = Parameter(test_value)
    np.testing.assert_array_equal(param.values, test_value)

def test_values():
    param = Parameter(test_value)
    with pytest.raises(ValueError):
        param.values = [[0, 5]]






from ezyrb import Snapshot
import numpy as np

import pytest

test_space = np.linspace(0, 3, 24)
test_value = test_space**2

def test_costructor():
    snap = Snapshot(test_value)
    snap = Snapshot(test_value, space=test_space)
    np.testing.assert_array_equal(snap.space, test_space)
    np.testing.assert_array_equal(snap.values, test_value)

def test_values():
    snap = Snapshot(test_value)
    snap.values = test_value
    snap = Snapshot(test_value, space=test_space)
    with pytest.raises(ValueError):
        snap.values = test_value[:-2]

def test_snapshot_space():
    snap = Snapshot(test_value)

def test_flattened():
    snap = Snapshot(test_value)
    assert snap.flattened.ndim == 1
    snap_3d = Snapshot(np.random.uniform(0, 1, size=(30, 3)))
    assert snap_3d.flattened.ndim == 1





""" Plugins submodule """

__all__ = [
    'Plugin',
    'DatabaseScaler',
    'ShiftSnapshots',
    'AutomaticShiftSnapshots',
    'Aggregation',
    'DatabaseSplitter',
    'DatabaseDictionarySplitter'
]

from .scaler import DatabaseScaler
from .plugin import Plugin
from .shift import ShiftSnapshots
from .automatic_shift import AutomaticShiftSnapshots
from .aggregation import Aggregation
from .database_splitter import DatabaseSplitter
from .database_splitter import DatabaseDictionarySplitter



import numpy as np

from unittest import TestCase
from ezyrb import Database


class TestDatabase(TestCase):
    def test_constructor_empty(self):
        a = Database()

    def test_constructor_arg(self):
        Database(np.random.uniform(size=(10, 3)),
                 np.random.uniform(size=(10, 8)))

    def test_constructor_arg_wrong(self):
        with self.assertRaises(ValueError):
            Database(np.random.uniform(size=(9, 3)),
                     np.random.uniform(size=(10, 8)))

    # def test_constructor_error(self):
    #     with self.assertRaises(TypeError):
    #         Database(np.eye(5))

    def test_getitem(self):
        org = Database(np.random.uniform(size=(10, 3)),
                       np.random.uniform(size=(10, 8)))
        new = org[2::2]
        assert new.parameters_matrix.shape[0] == new.snapshots_matrix.shape[0] == 4

    def test_getitem_singular(self):
        org = Database(np.random.uniform(size=(10, 3)),
                       np.random.uniform(size=(10, 8)))
        new = org[2]
        assert True

    def test_matrices(self):
        org = Database(np.random.uniform(size=(10, 3)),
                       np.random.uniform(size=(10, 8)))
        assert org.parameters_matrix.shape == (10, 3)
        assert org.snapshots_matrix.shape == (10, 8)

    def test_split(self):
        org = Database(np.random.uniform(size=(10, 3)),
                       np.random.uniform(size=(10, 8)))
        t1, t2 = org.split([8, 2])
        assert isinstance(t1, Database)
        assert isinstance(t2, Database)
        assert t1.parameters_matrix.shape == (8, org.parameters_matrix.shape[1])
        assert t2.parameters_matrix.shape == (2, org.parameters_matrix.shape[1])
        org = Database(np.random.uniform(size=(10, 3)),
                       np.random.uniform(size=(10, 8)))
        t1, t2, t3 = org.split([.3, .3, .4])
        assert isinstance(t1, Database)
        assert isinstance(t2, Database)
        assert isinstance(t3, Database)



"""Module for the snapshots database collected during the Offline stage."""

import numpy as np

from .parameter import Parameter
from .snapshot import Snapshot

class Database():
    """
    Database class

    :param array_like parameters: the input parameters
    :param array_like snapshots: the input snapshots
    :param Scale scaler_parameters: the scaler for the parameters. Default
        is None meaning no scaling.
    :param Scale scaler_snapshots: the scaler for the snapshots. Default is
        None meaning no scaling.
    :param array_like space: the input spatial data
    """
    def __init__(self, parameters=None, snapshots=None, space=None):
        self._pairs = []

        if parameters is None and snapshots is None:
            return

        if parameters is None:
            parameters = [None] * len(snapshots)
        elif snapshots is None:
            snapshots = [None] * len(parameters)

        if len(parameters) != len(snapshots):
            raise ValueError('parameters and snapshots must have the same length')

        for param, snap in zip(parameters, snapshots):
            param = Parameter(param)
            if isinstance(space, dict):
                snap_space = space.get(tuple(param.values), None)
                # print('snap_space', snap_space)
            else:
                snap_space = space
            snap = Snapshot(snap, space=snap_space)

            self.add(param, snap)

        # TODO: eventually improve the `space` assignment in the snapshots,
        # snapshots can have different space coordinates

    @property
    def parameters_matrix(self):
        """
        The matrix containing the input parameters (by row).

        :rtype: numpy.ndarray
        """
        return np.asarray([pair[0].values for pair in self._pairs])

    @property
    def snapshots_matrix(self):
        """
        The matrix containing the snapshots (by row).

        :rtype: numpy.ndarray
        """
        return np.asarray([pair[1].flattened for pair in self._pairs])

    def __getitem__(self, val):
        """
        This method returns a new Database with the selected parameters and
        snapshots.

        .. warning:: The new parameters and snapshots are a view of the
            original Database.
        """
        if isinstance(val, np.ndarray):
            view = Database()
            for p, s in np.asarray(self._pairs)[val]:
                view.add(p, s)
        elif isinstance(val, (int, slice)):
            view = Database()
            view._pairs = self._pairs[val]
        return view

    def __len__(self):
        """
        This method returns the number of snapshots.

        :rtype: int
        """
        return len(self._pairs)

    def __str__(self):
        """ Print minimal info about the Database """
        s = 'Database with {} snapshots and {} parameters'.format(
            self.snapshots_matrix.shape[1], self.parameters_matrix.shape[1])
        return s

    def add(self, parameter, snapshot):
        """
        Add (by row) new sets of snapshots and parameters to the original
        database.

        :param Parameter parameter: the parameter to add.
        :param Snapshot snapshot: the snapshot to add.
        """
        if not isinstance(parameter, Parameter):
            raise ValueError

        if not isinstance(snapshot, Snapshot):
            raise ValueError

        self._pairs.append((parameter, snapshot))

        return self


    def split(self, chunks, seed=None):
        """

        >>> db = Database(...)
        >>> train, test = db.split([0.8, 0.2]) # ratio
        >>> train, test = db.split([80, 20])   # n snapshots

        """

        if seed is not None:
            np.random.seed(seed)

        if all(isinstance(n, int) for n in chunks):
            if sum(chunks) != len(self):
                raise ValueError('chunk elements are inconsistent')

            ids = [
                j for j, chunk in enumerate(chunks)
                for i in range(chunk)
            ]
            np.random.shuffle(ids)


        elif all(isinstance(n, float) for n in chunks):
            if not np.isclose(sum(chunks), 1.):
                raise ValueError('chunk elements are inconsistent')


            cum_chunks = np.cumsum(chunks)
            cum_chunks = np.insert(cum_chunks, 0, 0.0)
            ids = np.ones(len(self)) * -1.
            tmp = np.random.uniform(0, 1, size=len(self))
            for i in range(len(cum_chunks)-1):
                is_between = np.logical_and(
                    tmp >= cum_chunks[i], tmp < cum_chunks[i+1])
                ids[is_between] = i

        else:
            ValueError

        new_database = [Database() for _ in range(len(chunks))]
        for i, chunk in enumerate(chunks):
            chunk_ids = np.array(ids) == i
            for p, s in np.asarray(self._pairs)[chunk_ids]:
                new_database[i].add(p, s)

        return new_database
    
    def get_snapshot_space(self, index):
        """
        Get the space coordinates of a snapshot by its index.

        :param int index: The index of the snapshot.
        :return: The space coordinates of the snapshot.
        :rtype: numpy.ndarray
        """
        if index < 0 or index >= len(self._pairs):
            raise IndexError("Snapshot index out of range.")
        return self._pairs[index][1].space



import numpy as np

from ezyrb import (GPR, Linear, RBF, ANN, KNeighborsRegressor,
    RadiusNeighborsRegressor)
import sklearn
import pytest

import torch.nn as nn
np.random.seed(17)

def get_xy():
    npts = 10
    dinput = 4

    inp = np.random.uniform(-1, 1, size=(npts, dinput))
    out = np.array([
        np.sin(inp[:, 0]) + np.sin(inp[:, 1]**2),
        np.cos(inp[:, 2]) + np.cos(inp[:, 3]**2)
    ]).T

    return inp, out

@pytest.mark.parametrize("model,kwargs", [
    (GPR, {}),
    (ANN, {'layers': [20, 20], 'function': nn.Tanh(), 'stop_training': 1e-8, 'last_identity': True}),
    (KNeighborsRegressor, {'n_neighbors': 1}),
    (RadiusNeighborsRegressor, {'radius': 0.1}),
    (Linear, {}),
])
class TestApproximation:
    def test_constructor_empty(self, model, kwargs):
        model = model(**kwargs)

    def test_fit(self, model, kwargs):
        x, y = get_xy()
        approx = model(**kwargs)
        approx.fit(x[:, 0].reshape(-1, 1), y[:, 0].reshape(-1, 1))

        approx = model(**kwargs)
        approx.fit(x, y)

    def test_predict_01(self, model, kwargs):
        x, y = get_xy()
        approx = model(**kwargs) 
        approx.fit(x, y)
        test_y = approx.predict(x)
        if isinstance(approx, ANN):
            np.testing.assert_array_almost_equal(y, test_y, decimal=3)
        else:
            np.testing.assert_array_almost_equal(y, test_y, decimal=6)


""" Module for Scaler plugin """

import numpy as np
import torch

from ezyrb import Database, Snapshot, Parameter
from .plugin import Plugin


class AutomaticShiftSnapshots(Plugin):
    """
    The plugin implements the automatic "shifting" preprocessing: exploiting a
    machine learning framework, it is able to detect the quantity to shift the
    snapshots composing the database, such that the reduction method performs
    better, depending on the problem at hand.

    Reference: Papapicco, D., Demo, N., Girfoglio, M., Stabile, G., & Rozza, G.
    (2022). The Neural Network shifted-proper orthogonal decomposition: A
    machine learning approach for non-linear reduction of hyperbolic equations.
    Computer Methods in Applied Mechanics and Engineering, 392, 114687.

    :param callable shift_function: a user defined function that return the
        shifting quantity for any the snapshot, given the corresponding input
        parameter.
    :param Approximation interpolator: the interpolator to use to evaluate the
        shifted snapshots on some reference space.
    :param int parameter_index: in case of multi-dimensional parameter,
        indicate the index of the parameter component to pass to the shift
        function. Default is 0.
    :param int reference_index: indicate the index of the snapshots within the
        database whose space will be used as reference space. Default is 0.

    Example:

    >>> from ezyrb import POD, RBF, Database, Snapshot, Parameter, Linear, ANN
    >>> from ezyrb import ReducedOrderModel as ROM
    >>> from ezyrb.plugin import AutomaticShiftSnapshots
    >>> interp = ANN([10, 10], torch.nn.Softplus(), 1000, frequency_print=50, lr=0.03)
    >>> shift = ANN([], torch.nn.LeakyReLU(), [2000, 1e-3], frequency_print=50, l2_regularization=0,  lr=0.002)
    >>> nnspod = AutomaticShiftSnapshots(shift, interp, Linear(fill_value=0.0), barycenter_loss=10.)
    >>> pod = POD(rank=1)
    >>> rbf = RBF()
    >>> db = Database()
    >>> for param in params:
    >>>     space, values = wave(param)
    >>>     snap = Snapshot(values=values, space=space)
    >>>     db.add(Parameter(param), snap)
    >>> rom = ROM(db, pod, rbf, plugins=[nnspod])
    >>> rom.fit()
    """
    def __init__(self, shift_network, interp_network, interpolator,
                 parameter_index=0, reference_index=0, barycenter_loss=0):
        super().__init__()

        self.interpolator = interpolator
        self.shift_network = shift_network
        self.interp_network = interp_network
        self.parameter_index = parameter_index
        self.reference_index = reference_index
        self.barycenter_loss = barycenter_loss

    def _train_interp_network(self):
        """
        """
        self.interp_network.fit(
            self.reference_snapshot.space.reshape(-1, 1),
            self.reference_snapshot.values.reshape(-1, 1)
        )

    def _train_shift_network(self, db):
        """
        """
        ref_center = torch.tensor(np.average(
            self.reference_snapshot.space * self.reference_snapshot.values))

        input_ = torch.from_numpy(np.vstack([
            np.vstack([s.space, np.ones(shape=(s.space.shape[0],))*p.values]).T
            for p, s in db._pairs 
        ])).float()
        output_ = torch.from_numpy(np.vstack([
            self.reference_snapshot.space.reshape(-1, 1)
            for _ in db._pairs
        ]))

        self.shift_network._build_model(input_, output_)
        optimizer = self.shift_network.optimizer(
            self.shift_network.model.parameters(),
            lr=self.shift_network.lr,
            weight_decay=self.shift_network.l2_regularization)

        n_epoch = 1
        flag = True
        while flag:

            shifts = self.shift_network.model(input_).float()
            loss = torch.tensor([0.0])
            for (_, snap), shift in zip(db._pairs, np.split(shifts, len(db))):

                tensor_space = torch.from_numpy(snap.space)
                tensor_values = torch.from_numpy(snap.values)

                translated_space = tensor_space - shift.reshape(snap.space.shape)
                translated_space = translated_space.float()
                interpolated_reference_values = self.interp_network.model(translated_space.reshape(-1, 1)).float().flatten()

                diff = torch.mean(
                    (tensor_values - interpolated_reference_values)**2)

                if self.barycenter_loss:
                    snap_center = torch.mean(translated_space * tensor_values)
                    diff += self.barycenter_loss*(ref_center - snap_center)**2

                loss += diff

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            scalar_loss = loss.item()
            self.shift_network.loss_trend.append(scalar_loss)

            for criteria in self.shift_network.stop_training:
                if isinstance(criteria, int):  # stop criteria is an integer
                    if n_epoch == criteria:
                        flag = False
                elif isinstance(criteria, float):  # stop criteria is float
                    if scalar_loss < criteria:
                        flag = False

            if (flag is False or
                    n_epoch == 1 or n_epoch % self.shift_network.frequency_print == 0):
                print(f'[epoch {n_epoch:6d}]\t{scalar_loss:e}')

            n_epoch += 1

    def fit_preprocessing(self, rom):
        db = rom.database

        reference_snapshot = db._pairs[self.reference_index][1]
        self.reference_snapshot = reference_snapshot

        self._train_interp_network()
        self._train_shift_network(db)

        for param, snap in db._pairs:
            input_shift = np.hstack([
                snap.space.reshape(-1, 1),
                np.ones(shape=(snap.space.shape[0], 1))*param.values])
            shift = self.shift_network.predict(input_shift)
            
            self.interpolator.fit(
                snap.space.reshape(-1, 1) - shift,
                snap.values.reshape(-1, 1))

            snap.values = self.interpolator.predict(
                reference_snapshot.space.reshape(-1, 1)).flatten()  # reconstructing shifted snapshots in physical space 

    def predict_postprocessing(self, rom):
        
        ref_space = self.reference_snapshot.space
        db = Database()

        for param, snap in rom.predict_full_database._pairs:
            input_shift = np.hstack([
                ref_space.reshape(-1, 1),
                np.ones(shape=(ref_space.shape[0], 1))*param.values])
            shift = self.shift_network.predict(input_shift)
            snap.space = ref_space + shift.flatten()    # shifted space transports to correct physical frame 
            snap.space = snap.space.flatten()

            self.interpolator.fit(snap.space, snap.values.reshape(-1, 1))
            snap.values = self.interpolator.predict(ref_space)  # reconstruct snapshot in physical space

            snaps = Snapshot(values = snap.values, space = ref_space)
            db.add(Parameter(param.values), snaps)

        rom._full_database = db


import numpy as np
import pytest

from ezyrb import POD, GPR, RBF, Database, ANN
from ezyrb import KNeighborsRegressor, RadiusNeighborsRegressor, Linear
from ezyrb import ReducedOrderModel as ROM
from ezyrb.plugin.scaler import DatabaseScaler

from sklearn.preprocessing import StandardScaler, MinMaxScaler

snapshots = np.load('tests/test_datasets/p_snapshots.npy').T
pred_sol_tst = np.load('tests/test_datasets/p_predsol.npy').T
pred_sol_gpr = np.load('tests/test_datasets/p_predsol_gpr.npy').T
param = np.array([[-.5, -.5], [.5, -.5], [.5, .5], [-.5, .5]])


def test_constructor():
    pod = POD()
    import torch
    rbf = RBF()
    #rbf = ANN([10, 10], function=torch.nn.Softplus(), stop_training=[1000])
    db = Database(param, snapshots.T)
    # rom = ROM(db, pod, rbf, plugins=[DatabaseScaler(StandardScaler(), 'full', 'snapshots')])
    rom = ROM(db, pod, rbf, plugins=[
        DatabaseScaler(StandardScaler(), 'reduced', 'parameters'),
        DatabaseScaler(StandardScaler(), 'reduced', 'snapshots')
    ])
    rom.fit()
   
    


def test_values():
    pod = POD()
    rbf = RBF()
    db = Database(param, snapshots.T)
    rom = ROM(db, pod, rbf, plugins=[
        DatabaseScaler(StandardScaler(), 'reduced', 'snapshots'),
        DatabaseScaler(StandardScaler(), 'full', 'parameters')
    ])
    rom.fit()
    test_param = param[2]
    truth_sol = db.snapshots_matrix[2]
    predicted_sol = rom.predict(test_param)[0]
    np.testing.assert_allclose(predicted_sol, truth_sol, 
            rtol=1e-5, atol=1e-5)




""" Module for Scaler plugin """

from .plugin import Plugin


class ShiftSnapshots(Plugin):
    """
    The plugin implements the "shifting" preprocessing: it makes possible the
    rigid shift of the snapshots composing the database, such that the
    reduction method performs better, dipendentely by the problem at hand.
    The shift function has to be passed by the user, together with an
    `Approximation` class in order to evaluate the translated snapshots onto
    the space of a custom reference space.

    Reference: Reiss, J., Schulze, P., Sesterhenn, J., & Mehrmann, V. (2018).
    The shifted proper orthogonal decomposition: A mode decomposition for
    multiple transport phenomena. SIAM Journal on Scientific Computing, 40(3),
    A1322-A1344.

    :param callable shift_function: a user defined function that return the
        shifting quantity for any the snapshot, given the corresponding input
        parameter.
    :param Approximation interpolator: the interpolator to use to evaluate the
        shifted snapshots on some reference space.
    :param int parameter_index: in case of multi-dimensional parameter,
        indicate the index of the parameter component to pass to the shift
        function. Default is 0.
    :param int reference_index: indicate the index of the snapshots within the
        database whose space will be used as reference space. Default is 0.

    Example:


    >>> def shift(time):
    >>>     return time-0.5
    >>> pod = POD()
    >>> rbf = RBF()
    >>> db = Database()
    >>> for param in params:
    >>>     space, values = wave(param)
    >>>     snap = Snapshot(values=values, space=space)
    >>>     db.add(Parameter(param), snap)
    >>> rom = ROM(db, pod, rbf, plugins=[ShiftSnapshots(shift, RBF())])
    >>> rom.fit()

    """
    def __init__(self, shift_function, interpolator, parameter_index=0,
                 reference_index=0):
        super().__init__()

        self.__shift_function = shift_function
        self.interpolator = interpolator
        self.parameter_index = parameter_index
        self.reference_index = reference_index

    def fit_preprocessing(self, rom):
        db = rom.database

        reference_snapshot = db._pairs[self.reference_index][1]

        for param, snap in db._pairs:
            snap.space = reference_snapshot.space
            shift = self.__shift_function(param.values[self.parameter_index])
            self.interpolator.fit(
                snap.space.reshape(-1, 1) - shift,
                snap.values.reshape(-1, 1))

            snap.values = self.interpolator.predict(
                reference_snapshot.space.reshape(-1, 1)).flatten()

        rom.database = db

    def predict_postprocessing(self, rom):
        for param, snap in rom.predicted_full_database._pairs:
            snap.space = (
                rom.database._pairs[self.reference_index][1].space +
                self.__shift_function(param.values)
            )



"""EZyRB package"""

__all__ = [
    'Approximation', 'RBF', 'Linear', 'GPR',
    'ANN', 'KNeighborsRegressor', 'RadiusNeighborsRegressor'
]

from .approximation import Approximation
from .rbf import RBF
from .linear import Linear
from .gpr import GPR
from .ann import ANN
from .kneighbors_regressor import KNeighborsRegressor
from .radius_neighbors_regressor import RadiusNeighborsRegressor



import numpy as np

from unittest import TestCase
from ezyrb import POD, GPR, RBF, Database
from ezyrb import KNeighborsRegressor, RadiusNeighborsRegressor, Linear
from ezyrb import ReducedOrderModel as ROM
from ezyrb.reducedordermodel import MultiReducedOrderModel as MROM

snapshots = np.load('tests/test_datasets/p_snapshots.npy').T
pred_sol_tst = np.load('tests/test_datasets/p_predsol.npy').T
pred_sol_gpr = np.load('tests/test_datasets/p_predsol_gpr.npy').T
param = np.array([[-.5, -.5], [.5, -.5], [.5, .5], [-.5, .5]])


class TestReducedOrderModel(TestCase):
    def test_constructor(self):
        pod = POD()
        rbf = RBF()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, rbf)

    def test_fit(self):
        pod = POD()
        rbf = RBF()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, rbf).fit()

    def test_save(self):
        fname = 'ezyrb.tmp'
        pod = POD(rank=2)
        rbf = RBF()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, rbf)
        rom.fit()
        rom.save(fname)

    def test_load(self):
        fname = 'ezyrb.tmp'
        pod = POD()
        rbf = RBF()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, rbf)
        rom.fit()
        rom.save(fname)
        new_rom = ROM.load(fname)
        new_param = [-0.293344, -0.23120537]
        np.testing.assert_array_almost_equal(
            rom.predict(new_param),
            new_rom.predict(new_param)
        )

    def test_load2(self):
        fname = 'ezyrb2.tmp'
        pod = POD()
        rbf = RBF()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, rbf)
        rom.fit()
        rom.save(fname, save_db=False)
        new_rom = ROM.load(fname)
        new_param = [-0.293344, -0.23120537]
        np.testing.assert_array_almost_equal(
            rom.predict(new_param),
            new_rom.predict(new_param)
        )

    def test_predict_01(self):
        pod = POD()
        rbf = RBF()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, rbf).fit()
        pred_sol = rom.predict([-0.293344, -0.23120537])
        np.testing.assert_allclose(
            pred_sol.flatten(),
            pred_sol_tst, rtol=1e-4, atol=1e-5)

    def test_predict_02(self):
        np.random.seed(117)
        pod = POD(method='svd', rank=4)
        gpr = GPR()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, gpr).fit()
        pred_sol = rom.predict([-.45, -.45])
        np.testing.assert_allclose(
            pred_sol.flatten(),
            pred_sol_gpr, rtol=1e-4, atol=1e-5)

    def test_predict_03(self):
        pod = POD(method='svd', rank=3)
        gpr = GPR()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, gpr).fit()
        pred_sol = rom.predict(db.parameters_matrix[2])
        assert pred_sol[0].shape == db.snapshots_matrix[0].shape
        pred_db = rom.predict(db)
        assert pred_db.snapshots_matrix.shape == db.snapshots_matrix.shape

    def test_predict_04(self):
        pod = POD(method='svd', rank=3)
        gpr = GPR()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, gpr).fit()
        pred_sol = rom.predict(db.parameters_matrix)
        assert pred_sol.shape == db.snapshots_matrix.shape

    # def test_predict_scaler_01(self):
    #     from sklearn.preprocessing import StandardScaler
    #     scaler = StandardScaler()
    #     pod = POD()
    #     rbf = RBF()
    #     db = Database(param, snapshots.T, scaler_snapshots=scaler)
    #     rom = ROM(db, pod, rbf).fit()
    #     pred_sol = rom.predict(db.parameters[0])
    #     np.testing.assert_allclose(pred_sol, db._snapshots[0], rtol=1e-4, atol=1e-5)
    #     pred_sol = rom.predict(db.parameters[0:2])
    #     np.testing.assert_allclose(pred_sol, db._snapshots[0:2], rtol=1e-4, atol=1e-5)

    # def test_predict_scaler_02(self):
    #     from sklearn.preprocessing import StandardScaler
    #     scaler_p = StandardScaler()
    #     scaler_s = StandardScaler()
    #     pod = POD()
    #     rbf = RBF()
    #     db = Database(param, snapshots.T, scaler_parameters=scaler_p, scaler_snapshots=scaler_s)
    #     rom = ROM(db, pod, rbf).fit()
    #     pred_sol = rom.predict(db._parameters[0])
    #     np.testing.assert_allclose(pred_sol, db._snapshots[0], rtol=1e-4, atol=1e-5)
    #     pred_sol = rom.predict(db._parameters[0:2])
    #     np.testing.assert_allclose(pred_sol, db._snapshots[0:2], rtol=1e-4, atol=1e-5)

    # def test_predict_scaling_coeffs(self):
    #     from sklearn.preprocessing import StandardScaler
    #     scaler = StandardScaler()
    #     pod = POD()
    #     rbf = RBF()
    #     db = Database(param, snapshots.T)
    #     rom = ROM(db, pod, rbf, scaler).fit()
    #     pred_sol = rom.predict(db._parameters[0])
    #     np.testing.assert_allclose(pred_sol, db._snapshots[0], rtol=1e-4, atol=1e-5)
    #     pred_sol = rom.predict(db._parameters[0:2])
    #     np.testing.assert_allclose(pred_sol, db._snapshots[0:2], rtol=1e-4, atol=1e-5)

    def test_test_error(self):
        pod = POD(method='svd', rank=-1)
        rbf = RBF()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, rbf).fit()
        error = rom.test_error(db)
        np.testing.assert_almost_equal(error, 0, decimal=6)

    def test_kfold_cv_error_01(self):
        pod = POD()
        rbf = RBF()
        db = Database(param, snapshots.T)
        n_splits = len(db)
        rom = ROM(db, pod, rbf)
        err_kfold = rom.kfold_cv_error(n_splits=n_splits)
        err_loo = rom.loo_error()
        np.testing.assert_allclose(err_kfold, err_loo)

    def test_loo_error_01(self):
        pod = POD()
        rbf = RBF()
        gpr = GPR()
        rnr = RadiusNeighborsRegressor()
        knr = KNeighborsRegressor(n_neighbors=1)
        lin = Linear(fill_value=0)
        db = Database(param, snapshots.T)
        exact_len = len(db)
        approximations = [rbf, gpr, knr, rnr]#, lin]
        roms = [ROM(db, pod, app) for app in approximations]
        len_errors = [len(rom.loo_error()) for rom in roms]
        np.testing.assert_allclose(len_errors, exact_len)

    def test_loo_error_02(self):
        pod = POD()
        gpr = GPR()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, gpr)
        err = rom.loo_error()
        np.testing.assert_allclose(
            err[0],
            np.array(0.595857),
            rtol=1e-3)

    def test_loo_error_singular_values(self):
        pod = POD()
        rbf = RBF()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, rbf).fit()
        valid_svalues = rom.reduction.singular_values
        rom.loo_error()
        np.testing.assert_allclose(valid_svalues, rom.reduction.singular_values)

    def test_multi_db(self):
        pod = POD()
        pod2 = POD(rank=1)
        gpr = GPR()
        db1 = Database(param, snapshots.T)
        rom = MROM({'p': db1}, {'a': pod, 'b':pod2}, gpr).fit()
        pred = rom.predict([-.5, -.5])
        assert isinstance(pred, dict)
        assert len(pred) == 2


"""
    def test_optimal_mu(self):
        pod = POD()
        rbf = RBF()
        gpr = GPR()
        rnr = RadiusNeighborsRegressor()
        knr = KNeighborsRegressor(n_neighbors=1)
        lin = Linear()
        db = Database(param, snapshots.T)
        exact_len = param.shape[1]
        approximations = [rbf, gpr, knr, rnr, lin]
        for k in [1, 2]:
            roms = [ROM(db, pod, app).fit() for app in approximations]
            len_opt_mu = [rom.optimal_mu(k=k).shape[1] for rom in roms]
            np.testing.assert_allclose(len_opt_mu, exact_len)
            len_k = [rom.optimal_mu(k=k).shape[0] for rom in roms]
            np.testing.assert_allclose(len_k, k)
"""


""" Module for discretized solution object """

import numpy as np
import matplotlib.pyplot as plt


class Snapshot:

    def __init__(self, values, space=None):
        if isinstance(values, Snapshot):
            self.values = values.values
            self.space = values.space
        else:
            self.values = values
            self.space = space

    @property
    def values(self):
        """ 
        Get the snapshot values.
        """
        return self._values

    @values.setter
    def values(self, new_values):
        if hasattr(self, 'space') and self.space is not None:
            if len(self.space) != len(new_values):
                raise ValueError('invalid ndof for the current space.')

        self._values = new_values

    @property
    def space(self):
        """ 
        Get the snapshot space.
        """
        return self._space

    @space.setter
    def space(self, new_space):
        if hasattr(self, 'values') and self.values is not None:
            if new_space is not None and len(self.values) != len(new_space):
                raise ValueError('invalid ndof for the current space.')

        self._space = new_space

    @property
    def flattened(self):
        """ return the values in 1D array """
        return self.values.flatten()

    def plot(self):
        """ Plot the snapshot, if possible. """

        if self.space is None:
            print('No space set, unable to plot.')
            return

        if np.asarray(self.space).ndim == 1:
            plt.plot(self.space, self.values)
        else:
            raise NotImplementedError



""" Module for Scaler plugin """

from .plugin import Plugin


class DatabaseScaler(Plugin):
    """
    The plugin to rescale the database of the reduced order model. It uses a
    user defined `scaler`, which has to have implemented the `fit`, `transform`
    and  `inverse_trasform` methods (i.e. `sklearn` interface), to rescale
    the parameters and/or the snapshots. It can be applied at the full order
    (`mode='full'`), at the reduced one (`mode='reduced'`) or both of them
    (`mode='both'`).

    :param obj scaler: a generic object which has to have implemented the
        `fit`, `transform` and  `inverse_trasform` methods (i.e. `sklearn`
        interface).
    :param {'full', 'reduced'} mode: define if the rescaling has to be
        applied at the full order ('full') or at the reduced one ('reduced').
    :param {'parameters', 'snapshots'} params: define if the rescaling has to
        be applied to the parameters or to the snapshots.
    """
    def __init__(self, scaler, mode, target) -> None:
        super().__init__()

        self.scaler = scaler
        self.mode = mode
        self.target = target
        
    @property
    def target(self):
        """
        Get the type of scaling. See class documentation for more info.

        rtype: str
        """
        return self._target

    @target.setter
    def target(self, new_target):
        if new_target not in ['snapshots', 'parameters']:
            raise ValueError

        self._target = new_target

    @property
    def mode(self):
        """
        Get the type of scaling. See class documentation for more info.

        rtype: str
        """
        return self._mode

    @mode.setter
    def mode(self, new_mode):
        if new_mode not in ['full', 'reduced']:
            raise ValueError

        self._mode = new_mode

    def _select_matrix(self, db):
        """ Helper function to select the proper matrix to rescale. """
        return getattr(db, f'{self.target}_matrix')

    def rom_preprocessing(self, rom):
        if self.mode != 'reduced':
            return

        db = rom._reduced_database

        self.scaler.fit(self._select_matrix(db))

        if self.target == 'parameters':
            new_db = type(db)(
                self.scaler.transform(self._select_matrix(db)),
                db.snapshots_matrix
            )
        else:
            new_db = type(db)(
                db.parameters_matrix,
                self.scaler.transform(self._select_matrix(db)),
            )

        rom._reduced_database = new_db

    def fom_preprocessing(self, rom):
        if self.mode != 'full':
            return

        db = rom._full_database

        self.scaler.fit(self._select_matrix(db))

        if self.target == 'parameters':
            new_db = type(db)(
                self.scaler.transform(self._select_matrix(db)),
                db.snapshots_matrix
            )
        else:
            new_db = type(db)(
                db.parameters_matrix,
                self.scaler.transform(self._select_matrix(db)),
            )

        rom._full_database = new_db

    def fom_postprocessing(self, rom):

        if self.mode != 'full':
            return

        db = rom._full_database

        if self.target == 'parameters':
            new_db = type(db)(
                self.scaler.inverse_transform(self._select_matrix(db)),
                db.snapshots_matrix
            )
        else:
            new_db = type(db)(
                db.parameters_matrix,
                self.scaler.inverse_transform(self._select_matrix(db)),
            )

        rom._full_database = new_db

    def rom_postprocessing(self, rom):
        if self.mode != 'reduced':
            return

        db = rom._reduced_database

        if self.target == 'parameters':
            new_db = type(db)(
                self.scaler.inverse_transform(self._select_matrix(db)),
                db.snapshots_matrix
            )
        else:
            new_db = type(db)(
                db.parameters_matrix,
                self.scaler.inverse_transform(self._select_matrix(db)),
            )
            
        rom._reduced_database = new_db




from .plugin import Plugin
from ..database import Database


class DatabaseSplitter(Plugin):


    def __init__(self, train=0.9, test=0.1, validation=0.0, predict=0.0,
            seed=None):
        super().__init__()

        self.train = train
        self.test = test
        self.validation = validation
        self.predict = predict
        self.seed = seed

    def fit_preprocessing(self, rom):
        db = rom._database
        if isinstance(db, Database):
            train, test, validation, predict = db.split(
                [self.train, self.test, self.validation, self.predict],
                seed=self.seed
            )

        elif isinstance(db, dict):
            train, test, validation, predict = list(db.values())[0].split(
                [self.train, self.test, self.validation, self.predict],
                seed=self.seed
            )
            # TODO improve this splitting if needed (now only reading the database of
            # the first ROM)


        rom.train_full_database = train
        rom.test_full_database = test
        rom.validation_full_database = validation
        rom.predict_full_database = predict
        #print('train', train.snapshots_matrix.shape)
        #print('test', test.snapshots_matrix.shape)
        #print('validation', validation.snapshots_matrix.shape)
        #print('predict', predict.snapshots_matrix.shape)

class DatabaseDictionarySplitter(Plugin):

    """
    This plugin class is used to define the train, test, validation and predict
    databases when the databases are already split: train, test, validation and
    predict are already database objects stored in a dictionary. Given the desired keys
    of the dictionary as input, the plugin will assign the corresponding database
    objects to the train, test, validation and predict attributes of the ROM.
    """
    
    def __init__(self, train_key=None, test_key=None, validation_key=None,
                 predict_key=None):
        super().__init__()
        self.train_key = train_key
        self.test_key = test_key
        self.validation_key = validation_key
        self.predict_key = predict_key

    def fit_preprocessing(self, rom):
        db = rom._database
        if isinstance(db, dict):
            if self.train_key is not None:
                rom.train_full_database = db[self.train_key]
            if self.test_key is not None:
                rom.test_full_database = db[self.test_key]
            if self.validation_key is not None:
                rom.validation_full_database = db[self.validation_key]
            if self.predict_key is not None:
                rom.predict_full_database = db[self.predict_key]
        else:
            raise ValueError("The database must be a dictionary of databases.")
           


"""Module for the Reduced Order Modeling."""

import math
import copy
import pickle
import numpy as np
from scipy.spatial import Delaunay
from sklearn.model_selection import KFold
from .database import Database
from .reduction import Reduction
from .approximation import Approximation


from abc import ABC, abstractmethod

class ReducedOrderModelInterface(ABC):

    def _execute_plugins(self, when):
        """
        Execute the plugins at the specified time, if the plugin has the
        specified method.

        :param str when: the time when the plugins have to be executed.
            The available times are:
                - 'fit_preprocessing'
                - 'fit_before_reduction'
                - 'fit_after_reduction'
                - 'fit_before_approximation'
                - 'fit_after_approximation'
                - 'fit_postprocessing'
                - 'predict_preprocessing'
                - 'predict_postprocessing'
        """
        for plugin in self.plugins:
            if hasattr(plugin, when):
                getattr(plugin, when)(self)


class ReducedOrderModel(ReducedOrderModelInterface):
    """
    Reduced Order Model class.

    This class performs the actual reduced order model using the selected
    methods for approximation and reduction.

    :param ezyrb.Database database: the database to use for training the
        reduced order model.
    :param ezyrb.Reduction reduction: the reduction method to use in reduced
        order model.
    :param ezyrb.Approximation approximation: the approximation method to use
        in reduced order model.
    :param list plugins: list of plugins to use in the reduced order model.

    :cvar ezyrb.Database database: the database used for training the reduced
        order model.
    :cvar ezyrb.Reduction reduction: the reduction method used in reduced order
        model.
    :cvar ezyrb.Approximation approximation: the approximation method used in
        reduced order model.
    :cvar list plugins: list of plugins used in the reduced order model.

    :Example:

         >>> from ezyrb import ReducedOrderModel as ROM
         >>> from ezyrb import POD, RBF, Database
         >>> pod = POD()
         >>> rbf = RBF()
         >>> # param, snapshots and new_param are assumed to be declared
         >>> db = Database(param, snapshots)
         >>> rom = ROM(db, pod, rbf).fit()
         >>> rom.predict(new_param)

    """
    def __init__(self, database, reduction, approximation,
                 plugins=None):

        self.database = database
        self.reduction = reduction
        self.approximation = approximation

        if plugins is None:
            plugins = []

        self.plugins = plugins

        self.clean()

    def clean(self):
        self.train_full_database = None
        self.train_reduced_database = None
        self.predict_full_database = None
        self.predict_reduced_database = None
        self.test_full_database = None
        self.test_reduced_database = None
        self.validation_full_database = None
        self.validation_reduced_database = None

    @property
    def database(self):
        return self._database

    @database.setter
    def database(self, value):

        if not isinstance(value, (Database, dict)):
            raise TypeError(
                "The database has to be an instance of the Database class, or a dictionary of Database.")

        self._database = value

    @database.deleter
    def database(self):
        del self._database

    @property
    def reduction(self):
        return self._reduction

    @reduction.setter
    def reduction(self, value):
        if not isinstance(value, Reduction):
            raise TypeError(
                "The reduction has to be an instance of the Reduction class, or a dict of Reduction.")

        self._reduction = value

    @reduction.deleter
    def reduction(self):
        del self._reduction

    @property
    def approximation(self):
        return self._approximation

    @approximation.setter
    def approximation(self, value):
        if not isinstance(value, Approximation):
            raise TypeError(
                "The approximation has to be an instance of the Approximation class, or a dict of Approximation.")

        self._approximation = value

    @approximation.deleter
    def approximation(self):
        del self._approximation

    @property
    def n_database(self):
        value_, class_ = self.database, Database
        return len(value_) if not isinstance(value_, class_) else 1

    @property
    def n_reduction(self):
        value_, class_ = self.reduction, Reduction
        return len(value_) if not isinstance(value_, class_) else 1

    @property
    def n_approximation(self):
        value_, class_ = self.approximation, Approximation
        return len(value_) if isinstance(value_, class_) else 1

    def fit_reduction(self):

        # for k, rom_ in self.roms.items():
        #     rom_['reduction'].fit(rom_['database'].snapshots_matrix.T)
        if not hasattr(self, 'train_full_database'):
            raise RuntimeError

        self.reduction.fit(self.train_full_database.snapshots_matrix.T)

    def _reduce_database(self, db):
        return Database(
            db.parameters_matrix,
            self.reduction.transform(db.snapshots_matrix.T).T
        )

    def fit_approximation(self):

        if not hasattr(self, 'train_reduced_database'):
            raise RuntimeError

        self.approximation.fit(self.train_reduced_database.parameters_matrix,
                           self.train_reduced_database.snapshots_matrix)

        # if self.n_database == 1 and self.n_reduction == 1:
        #     self.train_full_database = self.database
        #     self.reduction.fit(self.database.snapshots_matrix.T)

        # elif self.n_database == 1 and self.n_reduction > 1:
        #     self.train_full_database = self.database
        #     for reduction in self.reduction:
        #         reduction.fit(self.database.snapshots_matrix.T)

        # elif self.n_database > 1 and self.n_reduction == 1:
        #     self.train_full_database = self.database
        #     self.reduction = [
        #         (k, copy.deepcopy(self.reduction))
        #         for k in self.train_full_database
        #     ]
        #     print(self.reduction)
        #     for reduction, database in zip(self.reduction, self.train_full_database):
        #         self.reduction[reduction].fit(self.train_full_database[database].snapshots_matrix.T)

        # elif self.n_database > 1 and self.n_reduction > 1:
        #     raise NotImplementedError
        # else:
        #     raise RuntimeError

    def fit(self):
        r"""
        Calculate reduced space

        """
        self._execute_plugins('fit_preprocessing')
        if self.train_full_database is None:
           self.train_full_database = copy.deepcopy(self.database)

        self._execute_plugins('fit_before_reduction')

        self.fit_reduction()
        self.train_reduced_database = self._reduce_database(
            self.train_full_database)

        self._execute_plugins('fit_after_reduction')
        ## FULL-ORDER PREPROCESSING here
        #for plugin in self.plugins:
        #    plugin.fom_preprocessing(self)
            
        #self.shifted_database = self._full_database

        #self.reduction.fit(self._full_database.snapshots_matrix.T)
        #reduced_snapshots = self.reduction.transform(
        #    self._full_database.snapshots_matrix.T).T

        self._execute_plugins('fit_before_approximation')

        self.fit_approximation()

        self._execute_plugins('fit_after_approximation')

        self._execute_plugins('fit_postprocessing')

        return self

    def predict(self, parameters):
        """
        Calculate predicted solution for given parameters. If `parameters` is
        a 2D array, the function returns a 2D array of predicted solutions.
        If `parameters` is a Database, the function returns the database of
        predicted solutions.

        :return: the database containing all the predicted solution (with
            corresponding parameters).
        :rtype: Database
        """
        self._execute_plugins('predict_preprocessing')

        if isinstance(parameters, Database):
            self.predict_reduced_database = parameters

        elif isinstance(parameters, (list, np.ndarray, tuple)):

            parameters = np.atleast_2d(parameters)
            self.predict_reduced_database = Database(
                parameters=parameters,
                snapshots=[None] * len(parameters)
            )
        else:
            raise TypeError

        self._execute_plugins('predict_before_approximation')
        self.predict_reduced_database = Database(
            self.predict_reduced_database.parameters_matrix,
            self.approximation.predict(
                self.predict_reduced_database.parameters_matrix).reshape(
                    self.predict_reduced_database.parameters_matrix.shape[0], -1
                )
        )
        
        self._execute_plugins('predict_after_approximation')
        #mu = np.atleast_2d(mu)
        
        #for plugin in self.plugins:
        #    if plugin.target == 'parameters':
        #        mu = plugin.scaler.transform(mu)
            
 
        #self._reduced_database = Database(
        #        mu, np.atleast_2d(self.approximation.predict(mu)))

        self._execute_plugins('predict_before_expansion')

        self.predicted_full_database = Database(
            self.predict_reduced_database.parameters_matrix,
            self.reduction.inverse_transform(
                    self.predict_reduced_database.snapshots_matrix.T).T
        )
        self._execute_plugins('predict_after_expansion')

        self._execute_plugins('predict_postprocessing')

        if isinstance(parameters, Database):
            return self.predicted_full_database #predict_full_database
        else:
            return self.predicted_full_database.snapshots_matrix #predict_full_database


    def save(self, fname, save_db=True, save_reduction=True, save_approx=True):
        """
        Save the object to `fname` using the pickle module.

        :param str fname: the name of file where the reduced order model will
            be saved.
        :param bool save_db: Flag to select if the `Database` will be saved.
        :param bool save_reduction: Flag to select if the `Reduction` will be
            saved.
        :param bool save_approx: Flag to select if the `Approximation` will be
            saved.

        Example:

        >>> from ezyrb import ReducedOrderModel as ROM
        >>> rom = ROM(...) #  Construct here the rom
        >>> rom.fit()
        >>> rom.save('ezyrb.rom')
        """
        rom_to_store = copy.copy(self)

        if not save_db:
            del rom_to_store.database
        if not save_reduction:
            del rom_to_store.reduction
        if not save_approx:
            del rom_to_store.approximation

        with open(fname, 'wb') as output:
            pickle.dump(rom_to_store, output, pickle.HIGHEST_PROTOCOL)

    @staticmethod
    def load(fname):
        """
        Load the object from `fname` using the pickle module.

        :return: The `ReducedOrderModel` loaded

        Example:

        >>> from ezyrb import ReducedOrderModel as ROM
        >>> rom = ROM.load('ezyrb.rom')
        >>> rom.predict(new_param)
        """
        with open(fname, 'rb') as output:
            rom = pickle.load(output)

        return rom

    def test_error(self, test, norm=np.linalg.norm):
        """
        Compute the mean norm of the relative error vectors of predicted
        test snapshots.

        :param database.Database test: the input test database.
        :param function norm: the function used to assign at the vector of
            errors a float number. It has to take as input a 'numpy.ndarray'
            and returns a float. Default value is the L2 norm.
        :return: the mean L2 norm of the relative errors of the estimated
            test snapshots.
        :rtype: numpy.ndarray
        """
        predicted_test = self.predict(test.parameters_matrix)
        return np.mean(
            norm(predicted_test - test.snapshots_matrix,
            axis=1) / norm(test.snapshots_matrix, axis=1))

    def kfold_cv_error(self, n_splits, *args, norm=np.linalg.norm, **kwargs):
        r"""
        Split the database into k consecutive folds (no shuffling by default).
        Each fold is used once as a validation while the k - 1 remaining folds
        form the training set. If `n_splits` is equal to the number of
        snapshots this function is the same as `loo_error` but the error here
        is relative and not absolute.

        :param int n_splits: number of folds. Must be at least 2.
        :param function norm: function to apply to compute the relative error
            between the true snapshot and the predicted one.
            Default value is the L2 norm.
        :param \*args: additional parameters to pass to the `fit` method.
        :param \**kwargs: additional parameters to pass to the `fit` method.
        :return: the vector containing the errors corresponding to each fold.
        :rtype: numpy.ndarray
        """
        error = []
        kf = KFold(n_splits=n_splits)
        for train_index, test_index in kf.split(self.database):
            new_db = self.database[train_index]
            rom = type(self)(new_db, copy.deepcopy(self.reduction),
                             copy.deepcopy(self.approximation)).fit(
                                 *args, **kwargs)

            error.append(rom.test_error(self.database[test_index], norm))

        return np.array(error)

    def loo_error(self, *args, norm=np.linalg.norm, **kwargs):
        r"""
        Estimate the approximation error using *leave-one-out* strategy. The
        main idea is to create several reduced spaces by combining all the
        snapshots except one. The error vector is computed as the difference
        between the removed snapshot and the projection onto the properly
        reduced space. The procedure repeats for each snapshot in the database.
        The `norm` is applied on each vector of error to obtained a float
        number.

        :param function norm: the function used to assign at each vector of
            error a float number. It has to take as input a 'numpy.ndarray` and
            returns a float. Default value is the L2 norm.
        :param \*args: additional parameters to pass to the `fit` method.
        :param \**kwargs: additional parameters to pass to the `fit` method.
        :return: the vector that contains the errors estimated for all
            parametric points.
        :rtype: numpy.ndarray
        """
        error = np.zeros(len(self.database))
        db_range = list(range(len(self.database)))

        for j in db_range:
            indeces = np.array([True] * len(self.database))
            indeces[j] = False

            new_db = self.database[indeces]
            test_db = self.database[~indeces]
            rom = type(self)(new_db, copy.deepcopy(self.reduction),
                             copy.deepcopy(self.approximation)).fit()

            error[j] = rom.test_error(test_db)

        return error

    def optimal_mu(self, error=None, k=1):
        """
        Return the parametric points where new high-fidelity solutions have to
        be computed in order to globally reduce the estimated error. These
        points are the barycentric center of the region (simplex) with higher
        error.

        :param numpy.ndarray error: the estimated error evaluated for each
            snapshot; if error array is not passed, it is computed using
            :func:`loo_error` with the default function. Default value is None.
        :param int k: the number of optimal points to return. Default value is
            1.
        :return: the optimal points
        :rtype: numpy.ndarray
        """
        if error is None:
            error = self.loo_error()

        mu = self.database.parameters_matrix
        tria = Delaunay(mu)

        error_on_simplex = np.array([
            np.sum(error[smpx]) * self._simplex_volume(mu[smpx])
            for smpx in tria.simplices
        ])

        barycentric_point = []
        for index in np.argpartition(error_on_simplex, -k)[-k:]:
            worst_tria_pts = mu[tria.simplices[index]]
            worst_tria_err = error[tria.simplices[index]]

            barycentric_point.append(
                np.average(worst_tria_pts, axis=0, weights=worst_tria_err))

        return np.asarray(barycentric_point)

    def _simplex_volume(self, vertices):
        """
        Method implementing the computation of the volume of a N dimensional
        simplex.
        Source from: `wikipedia.org/wiki/Simplex
        <https://en.wikipedia.org/wiki/Simplex>`_.

        :param numpy.ndarray simplex_vertices: Nx3 array containing the
            parameter values representing the vertices of a simplex. N is the
            dimensionality of the parameters.
        :return: N dimensional volume of the simplex.
        :rtype: float
        """
        distance = np.transpose([vertices[0] - vi for vi in vertices[1:]])
        return np.abs(
            np.linalg.det(distance) / math.factorial(vertices.shape[1]))

class MultiReducedOrderModel(ReducedOrderModelInterface):
    """
    Multiple Reduced Order Model class.

    This class performs the actual reduced order model using the selected
    methods for approximation and reduction.

    :param ezyrb.Database database: the database to use for training the
        reduced order model.
    :param ezyrb.Reduction reduction: the reduction method to use in reduced
        order model.
    :param ezyrb.Approximation approximation: the approximation method to use
        in reduced order model.
    :param list plugins: list of plugins to use in the reduced order model.

    :cvar ezyrb.Database database: the database used for training the reduced
        order model.
    :cvar ezyrb.Reduction reduction: the reduction method used in reduced order
        model.
    :cvar ezyrb.Approximation approximation: the approximation method used in
        reduced order model.
    :cvar list plugins: list of plugins used in the reduced order model.

    :Example:

         >>> from ezyrb import ReducedOrderModel as ROM
         >>> from ezyrb import POD, RBF, Database
         >>> pod = POD()
         >>> rbf = RBF()
         >>> # param, snapshots and new_param are assumed to be declared
         >>> db = Database(param, snapshots)
         >>> rom = ROM(db, pod, rbf).fit()
         >>> rom.predict(new_param)

    """
    def __init__(self, *args, plugins=None, rom_plugin=None):

        if len(args) == 3:
            self.database = args[0]
            self.reduction = args[1]
            self.approximation = args[2]

            from itertools import product

            element_keys = product(
                self.database.keys(),
                self.reduction.keys(),
                self.approximation.keys()
            )
            self.roms = {

                tuple(key): ReducedOrderModel(
                    copy.deepcopy(self.database[key[0]]),
                    copy.deepcopy(self.reduction[key[1]]),
                    copy.deepcopy(self.approximation[key[2]])
                )
                for key in element_keys
            }
        elif (len(args) == 2 and isinstance(args[0], Database)
            and isinstance(args[1], dict)):
            self.database = args[0]
            self.roms = args[1]

        elif len(args) == 1 and isinstance(args[0], dict):
            self.roms = args[0]
            roms_keys = list(self.roms.keys())
            self.database = {roms_keys[i]: self.roms[roms_keys[i]].database
                        for i in range(len(self.roms))}

        if plugins is None:
            plugins = []
        self.plugins = plugins

        if rom_plugin is not None:
            for rom_ in self.roms.values():
                rom_.plugins.append(rom_plugin)


    @property
    def database(self):
        return self._database

    @database.setter
    def database(self, value):

        if not isinstance(value, (Database, dict)):
            raise TypeError(
                "The database has to be an instance of the Database class, or a dictionary of Database.")

        if isinstance(value, Database):
            self._database = {0: value}
        else:
            self._database = value

    @database.deleter
    def database(self):
        del self._database

    @property
    def reduction(self):
        return self._reduction

    @reduction.setter
    def reduction(self, value):
        if not isinstance(value, (Reduction, dict)):
            raise TypeError(
                "The reduction has to be an instance of the Reduction class, or a dict of Reduction.")

        if isinstance(value, Reduction):
            self._reduction = {0: value}
        else:
            self._reduction = value

    @reduction.deleter
    def reduction(self):
        del self._reduction

    @property
    def approximation(self):
        return self._approximation

    @approximation.setter
    def approximation(self, value):
        if not isinstance(value, (Approximation, dict)):
            raise TypeError(
                "The approximation has to be an instance of the Approximation class, or a dict of Approximation.")

        if isinstance(value, Approximation):
            self._approximation = {0: value}
        else:
            self._approximation = value

    @property
    def n_database(self):
        value_, class_ = self.database, Database
        return len(value_) if not isinstance(value_, class_) else 1

    @property
    def n_reduction(self):
        value_, class_ = self.reduction, Reduction
        return len(value_) if not isinstance(value_, class_) else 1

    @property
    def n_approximation(self):
        value_, class_ = self.approximation, Approximation
        return len(value_) if isinstance(value_, class_) else 1
        
    def fit(self):
        r"""
        Calculate reduced space
        """
        self._execute_plugins('fit_preprocessing')

        for rom_ in self.roms.values():
            if rom_.train_reduced_database==None:
                rom_.fit()
        
        self._execute_plugins('fit_postprocessing')
        return self

    

    def predict(self, parameters=None):
        """
        Calculate predicted solution for given mu
        If gaussians is True, the standard aggregation technique (gaussians) is used,
        otherwise the ANNs are used.

        :return: the database containing all the predicted solution (with
            corresponding parameters).
        :rtype: Database
        """

        self._execute_plugins('predict_preprocessing')

        # convert parameters from Database to numpy array (if database)
        if isinstance(parameters, Database):
            self.predict_reduced_database = parameters

        elif isinstance(parameters, (list, np.ndarray, tuple)):
            print(parameters)
            parameters = np.atleast_2d(parameters)
            self.predict_reduced_database = Database(parameters, [None]*len(parameters))
        elif parameters is None:
            if self.predict_full_database is None:
                raise RuntimeError
        else:
            raise TypeError

        self.multi_predict_database = {}
        for k, rom_ in self.roms.items():
            self.multi_predict_database[k] = rom_.predict(self.predict_reduced_database)
        print(self.multi_predict_database)
        self._execute_plugins('predict_postprocessing')

        if isinstance(parameters, Database):
            return self.multi_predict_database
        else:
            return {
                k:db.snapshots_matrix 
                for k, db in self.multi_predict_database.items()
            }
            

    def save(self, fname, save_db=True, save_reduction=True, save_approx=True):
        """
        Save the object to `fname` using the pickle module.

        :param str fname: the name of file where the reduced order model will
            be saved.
        :param bool save_db: Flag to select if the `Database` will be saved.
        :param bool save_reduction: Flag to select if the `Reduction` will be
            saved.
        :param bool save_approx: Flag to select if the `Approximation` will be
            saved.

        Example:

        >>> from ezyrb import ReducedOrderModel as ROM
        >>> rom = ROM(...) #  Construct here the rom
        >>> rom.fit()
        >>> rom.save('ezyrb.rom')
        """
        rom_to_store = copy.copy(self)

        if not save_db:
            del rom_to_store.database
        if not save_reduction:
            del rom_to_store.reduction
        if not save_approx:
            del rom_to_store.approximation

        with open(fname, 'wb') as output:
            pickle.dump(rom_to_store, output, pickle.HIGHEST_PROTOCOL)

    @staticmethod
    def load(fname):
        """
        Load the object from `fname` using the pickle module.

        :return: The `ReducedOrderModel` loaded

        Example:

        >>> from ezyrb import ReducedOrderModel as ROM
        >>> rom = ROM.load('ezyrb.rom')
        >>> rom.predict(new_param)
        """
        with open(fname, 'rb') as output:
            rom = pickle.load(output)

        return rom

    def test_error(self, test, norm=np.linalg.norm, relative=True):
        """
        Compute the mean norm of the relative error vectors of predicted
        test snapshots.

        :param database.Database test: the input test database.
        :param function norm: the function used to assign at the vector of
            errors a float number. It has to take as input a 'numpy.ndarray'
            and returns a float. Default value is the L2 norm.
        :param relative: True if the error computed is relative. Default is 
            True. 
        :return: the mean L2 norm of the relative errors of the estimated
            test snapshots.
        :rtype: numpy.ndarray
        """
        predicted_test = self.predict(test.parameters_matrix)
        if relative:
            return np.mean(
                norm(predicted_test.snapshots_matrix - test.snapshots_matrix,
                axis=1) / norm(test.snapshots_matrix, axis=1))
        else:
            return np.mean(
                norm(predicted_test.snapshots_matrix - test.snapshots_matrix,
                axis=1))

    def kfold_cv_error(self, n_splits, *args, norm=np.linalg.norm, relative=True, 
                       **kwargs):
        r"""
        Split the database into k consecutive folds (no shuffling by default).
        Each fold is used once as a validation while the k - 1 remaining folds
        form the training set. If `n_splits` is equal to the number of
        snapshots this function is the same as `loo_error` but the error here
        is relative and not absolute.

        :param int n_splits: number of folds. Must be at least 2.
        :param function norm: function to apply to compute the relative error
            between the true snapshot and the predicted one.
            Default value is the L2 norm.
        :param relative: True if the error computed is relative. Default is 
            True. 
        :param \*args: additional parameters to pass to the `fit` method.
        :param \**kwargs: additional parameters to pass to the `fit` method.
        :return: the vector containing the errors corresponding to each fold.
        :rtype: numpy.ndarray
        """
        error = []
        kf = KFold(n_splits=n_splits)
        for train_index, test_index in kf.split(self.database):
            new_db = self.database[train_index]
            rom = type(self)(new_db, copy.deepcopy(self.reduction),
                             copy.deepcopy(self.approximation)).fit(
                                 *args, **kwargs)

            error.append(rom.test_error(self.database[test_index], norm, relative))

        return np.array(error)

    def loo_error(self, *args, norm=np.linalg.norm, **kwargs):
        r"""
        Estimate the approximation error using *leave-one-out* strategy. The
        main idea is to create several reduced spaces by combining all the
        snapshots except one. The error vector is computed as the difference
        between the removed snapshot and the projection onto the properly
        reduced space. The procedure repeats for each snapshot in the database.
        The `norm` is applied on each vector of error to obtained a float
        number.

        :param function norm: the function used to assign at each vector of
            error a float number. It has to take as input a 'numpy.ndarray` and
            returns a float. Default value is the L2 norm.
        :param \*args: additional parameters to pass to the `fit` method.
        :param \**kwargs: additional parameters to pass to the `fit` method.
        :return: the vector that contains the errors estimated for all
            parametric points.
        :rtype: numpy.ndarray
        """
        error = np.zeros(len(self.database))
        db_range = list(range(len(self.database)))

        for j in db_range:
            indeces = np.array([True] * len(self.database))
            indeces[j] = False

            new_db = self.database[indeces]
            test_db = self.database[~indeces]
            rom = type(self)(new_db, copy.deepcopy(self.reduction),
                             copy.deepcopy(self.approximation)).fit()

            error[j] = rom.test_error(test_db)

        return error

    def optimal_mu(self, error=None, k=1):
        """
        Return the parametric points where new high-fidelity solutions have to
        be computed in order to globally reduce the estimated error. These
        points are the barycentric center of the region (simplex) with higher
        error.

        :param numpy.ndarray error: the estimated error evaluated for each
            snapshot; if error array is not passed, it is computed using
            :func:`loo_error` with the default function. Default value is None.
        :param int k: the number of optimal points to return. Default value is
            1.
        :return: the optimal points
        :rtype: numpy.ndarray
        """
        if error is None:
            error = self.loo_error()

        mu = self.database.parameters_matrix
        tria = Delaunay(mu)

        error_on_simplex = np.array([
            np.sum(error[smpx]) * self._simplex_volume(mu[smpx])
            for smpx in tria.simplices
        ])

        barycentric_point = []
        for index in np.argpartition(error_on_simplex, -k)[-k:]:
            worst_tria_pts = mu[tria.simplices[index]]
            worst_tria_err = error[tria.simplices[index]]

            barycentric_point.append(
                np.average(worst_tria_pts, axis=0, weights=worst_tria_err))

        return np.asarray(barycentric_point)

    def _simplex_volume(self, vertices):
        """
        Method implementing the computation of the volume of a N dimensional
        simplex.
        Source from: `wikipedia.org/wiki/Simplex
        <https://en.wikipedia.org/wiki/Simplex>`_.

        :param numpy.ndarray simplex_vertices: Nx3 array containing the
            parameter values representing the vertices of a simplex. N is the
            dimensionality of the parameters.
        :return: N dimensional volume of the simplex.
        :rtype: float
        """
        distance = np.transpose([vertices[0] - vi for vi in vertices[1:]])
        return np.abs(
            np.linalg.det(distance) / math.factorial(vertices.shape[1]))

    def reconstruction_error(self, db=None, relative=True, eps=1e-12):
        """
        Calculate the reconstruction error between the original snapshots and
        the ones reconstructed by the ROM.
        
        :param database.Database db: the database to use to compute the error.
            If None, the error is computed on the training database.
            Default is None.
        :param bool relative: True if the error computed is relative. Default is 
            True.
        :param float eps: small number to avoid division by zero in relative
            error computation. Default is 1e-12.
        :return: the vector containing the reconstruction errors.
        
        Esempio:    
        >>> from ezyrb import ReducedOrderModel as ROM
        >>> from ezyrb import POD, RBF, Database
        >>> db = Database(param, snapshots) # param and snapshots are assumed 
        to be declared
        >>> db_train = db[:10] # training database
        >>> db_test = db[10:] # test database
        >>> pod = POD()
        >>> rbf = RBF()
        >>> rom = ROM(db_train, pod, rbf)
        >>> rom.fit()
        >>> err_train_reduct = rom.reconstruction_error(relative=True)
        >>> err_test_reduct = rom.reconstruction_error(db_test, relative=True)
        """
        
        errs = []
        if db is None:
            db = self.database   
        snap = db.snapshots_matrix
        snap_red = self.reduction.transform(snap.T)
        snap_full = self.reduction.inverse_transform(snap_red).T

        E = snap - snap_full

        if relative:
            num = np.linalg.norm(E, axis=1)
            den = np.linalg.norm(snap, axis=1) + eps
            
            err = float(np.mean(num/den))        
        else:
            err = float(np.mean(np.linalg.norm(E, axis=1)))
        errs.append(err)
        
        return np.array(errs)

    def approximation_error(self, db=None, relative=True, eps=1e-12):
        """
        Calculate the approximation error between the true modal coefficients
        and the approximated ones.
        
        :param database.Database db: the database to use to compute the error.
            If None, the error is computed on the training database.
            Default is None.
        :param bool relative: True if the error computed is relative. Default is 
            True.
        :param float eps: small number to avoid division by zero in relative
            error computation. Default is 1e-12.
            
        :return: the vector containing the approximation errors.
        
        Esempio:
        >>> from ezyrb import ReducedOrderModel as ROM
        >>> from ezyrb import POD, RBF, Database
        >>> db = Database(param, snapshots) # param and snapshots are assumed 
        to be declared
        >>> db_train = db[:10] # training database
        >>> db_test = db[10:] # test database
        >>> pod = POD()
        >>> rbf = RBF()
        >>> rom = ROM(db_train, pod, rbf)
        >>> rom.fit()
        >>> err_train_approx = rom.approximation_error(relative=True)
        >>> err_test_approx = rom.approximation_error(db_test, relative=True)

        """
        errs = []
        if db is None:
            db = self.database

        snap = db.snapshots_matrix
        params_true = self.reduction.transform(snap.T).T

        params = db.parameters_matrix

        params_approx = self.approximation.predict(params)

        E = params_true - params_approx

        if relative:
            num = np.linalg.norm(E, axis=1)
            den = np.linalg.norm(params_true, axis=1) + eps

            err = float(np.mean(num/den))
        else:
            err = float(np.mean(np.linalg.norm(E, axis=1)))
        errs.append(err)

        return np.array(errs)

